/*
 * MdMDA - Md's MDA
 *
 * Copyright 1999 by Marco d'Itri <md@linux.it>
 *
 * You can use this software under the terms of the GPL. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 *
 * $Id: mdmda.c,v 1.2 1999/11/21 17:14:19 md Exp $
 */

#ifdef linux
#  define FSYNCDIR
#endif

/* How many levels of hashing? #undefine for none */
#define HASHED_SPOOL_DEPTH	3

/* Check quota? */
#define QUOTACHECK	3 * 1024 * 1024

#define MAILDIRSPOOL "/var/spool/mdmda"

/* System library */
#include <stdio.h>		/* fputs, printf, snprintf, perror */
#include <stdlib.h>		/* free */
#include <stdarg.h>		/* va_* */
#include <unistd.h>		/* getpid, fsync, close, write, unlink, stat */
#include <sys/stat.h>		/* mkdir, struct stat */
#include <fcntl.h>		/* open constants */
#ifdef QUOTACHECK
#include <dirent.h>
#endif
#include <time.h>		/* time */
#include <string.h>		/* strcat, strchr, strcpy, strdup, strsep */
#include <sysexits.h>		/* EX_TEMPFAIL 75 */
#include <limits.h>		/* PATH_MAX */
#include <sys/param.h>		/* MAXHOSTNAMELEN */
#include <errno.h>		/* ENOENT, EIO */

/* Function prototypes */
void err_quit(const char *,...);
void err_sys(const char *,...);
int mboxpath(char *, const char *);
int makemaildir(void);
int makeuserdir(const char *);
void maildir_append(const char *);
void mailbox_append(const char *);
#ifdef QUOTACHECK
int quotacheck(const char *, long);
long submaildirsize(char *);
#endif

int main(int argc, char *argv[])
{
    char mailbox[PATH_MAX + 30*3], user[30];

    /* check the arguments */
    if (argc < 2)
#ifdef QUOTACHECK
	err_quit("Usage: program username [quota]");
#else
	err_quit("Usage: program username");
#endif

    /* sanitize the username */
    if (strchr(argv[1], '/') || *argv[1] == '.')
	err_quit("Invalid character in username");
    if (snprintf(user, sizeof(user), "%s", argv[1]) == -1)
	err_quit("User name too long");

    /* chdir to the shared spool directory */
    if (chdir(MAILDIRSPOOL) < 0)
	err_sys("Cannot chdir to spool directory %s", MAILDIRSPOOL);

    /* make the mailbox name */
    if (mboxpath(mailbox, user) == -1)
	err_quit("Cannot determine mailbox path");

#ifdef QUOTACHECK
    /* check if the user is over quota */
    {
	long limit = 0;

	if (argc >= 3 && *argv[2])
	    limit = atol(argv[2]);
	if (!limit)
	    limit = QUOTACHECK;
	if (quotacheck(mailbox, QUOTACHECK)) {
	    fputs("Mailbox full.\n", stderr);
	    exit(EX_UNAVAILABLE);
	}
    }
#endif

    if (1)
	maildir_append(mailbox);
    else
	mailbox_append(mailbox);
    exit(0);
}

void maildir_append(const char *maildir)
{
    int fd;
    unsigned int n;
#ifdef QUOTACHECK
    unsigned long size = 0;
#endif
    char myhostname[MAXHOSTNAMELEN], file[MAXHOSTNAMELEN + 30], buf[BUFSIZ];
    char *file2;
    struct stat sb;
    time_t mytime;
    pid_t mypid = getpid();

    /* 1. chdir to the user maildir */
    while (chdir(maildir) < 0) {
	if (errno == ENOENT) {
	    if (makeuserdir(maildir) < 0)
		err_sys("Cannot create maildir %s", maildir);
	} else
	    err_sys("Cannot chdir to maildir %s", maildir);
    }

    /* 2. stat tmp/time.pid.host */
    gethostname(myhostname, sizeof(myhostname));
    for (n = 0;; n++) {
	mytime = time(NULL);
	sprintf(file, "tmp/%ld.%d.%s", (long)mytime, (int)mypid, myhostname);
	if (stat(file, &sb) < 0)
	    if (errno == ENOENT)
		break;		/* 3. OK */
	if (n > 2)
	    err_sys("Too many failed stats for file %s/%s", maildir, file);
	sleep(2);
    }

    /* 4. create tmp/time.pid.host */
    while ((fd = open(file, O_CREAT | O_EXCL | O_WRONLY, 0600)) < 0) {
	if (errno == ENOENT) {
	    if (makemaildir() < 0)
		err_sys("Cannot create maildir structure for %s", maildir);
	} else
	    err_sys("Cannot open file %s/%s", maildir, file);
    }

    /* 5. NFS-writes the message */
    while ((n = read(0, buf, BUFSIZ)) > 0) {
#ifdef QUOTACHECK
	size += n;
#endif
	if (write(fd, buf, n) != n) {
	    close(fd);
	    err_sys("Error writing %s", file);
	}
    }
    if (n < 0)
	err_sys("Error reading stdin (writing to %s/%s)", maildir, file);
    if ((fsync(fd)) < 0)
	err_sys("Error fsyncing %s", file);
    if ((close(fd)) < 0)
	err_sys("Error closing %s", file);

    /* 6. link the file to new/time.pid.host */
#ifdef QUOTACHECK
    file2 = malloc(MAXHOSTNAMELEN + 30 + 15);	/* 10 bytes for file size */
    sprintf(file2, "new%s,S=%ld", file + 3, size);
#else
    file2 = strdup(file);
    memcpy(file2, "new", 3);
#endif
    if (link(file, file2) < 0)
	err_sys("Error creating link %s", file2);
    /* We don't care about the result. Message has been delivered. */
    unlink(file);

#ifdef FSYNCDIR
    /* 6bis. update ext2fs directory metadata */
    if ((fd = open("new", O_RDONLY)) < 0)
	err_sys("Error opening directory for %s", file2);
    /* Linus wrote we should not care about other errors */
    if ((fsync(fd)) < 0 && errno == EIO)
	err_sys("Error fsyncing directory for %s", file2);
    close(fd);
#endif
    free(file2);
}

int makemaildir(void)
{
    const char *dirs[]={ "tmp", "new", "cur"  };
    int i;

    for (i = 0; i < 3; i++)
	if ((mkdir(dirs[i], 0700) < 0) && (errno != EEXIST))
	    return -1;
    return 0;
}

void mailbox_append(const char *mailbox)
{
}

int makeuserdir(const char *dir)
{
    char *ptr, *element, mydir[PATH_MAX];

    ptr = strdup(dir);
    while ((element = strsep(&ptr, "/"))) {
	strcat(mydir, element);
	strcat(mydir, "/");
	if ((mkdir(mydir, 0700) < 0) && (errno != EEXIST))
	    return -1;
    }
    free(ptr);
    return 0;
}

/* Determine the path of a mailbox */
int mboxpath(char *mailbox, const char *user)
{
#ifdef HASHED_SPOOL_DEPTH
    int n;
    const char *cp=user;

    for (n = 0; n < HASHED_SPOOL_DEPTH; n++) {
	if (*cp == 0)
	    break;
	*mailbox++ = *cp++;
	*mailbox++ = '/';
    }
#endif
    strcpy(mailbox, user);
    return 0;
}

#ifdef QUOTACHECK
/* check if the maildir is over quota */
int quotacheck(const char *maildir, long limit)
{
    long size;
    char *p;

    p = malloc(MAXHOSTNAMELEN + 30 + 4);
    strcpy(p, maildir);
    strcat(p, "/cur");
    size = submaildirsize(p);
    strcpy(p, maildir);
    strcat(p, "/new");
    size += submaildirsize(p);
    free(p);
    return size > limit ? 1 : 0;
}

long submaildirsize(char *dirname)
{
    long size = 0;
    DIR *dir;
    struct dirent *ep;
    char *p;

    if ((dir = opendir(dirname)) == NULL)
	err_sys("Error opening directory %s", p);
    while ((ep = readdir(dir)) != NULL) {
	p = ep->d_name;
	while (*p++) {
	    if (*p == ',' && *(p+1) == 'S' && *(p+2) == '=') {
		size += atoi(p+3);
		break;
	    }
	    if (!*p)
		break;
	}
    }
    closedir(dir);
    return size;
}
#endif

/* Error routines */
void err_sys(const char *fmt,...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, ": %s\n", strerror(errno));
    va_end(ap);
    exit(EX_TEMPFAIL);
}

void err_quit(const char *fmt,...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fputs("\n", stderr);
    va_end(ap);
    exit(EX_TEMPFAIL);
}

